import java.io.*;

/**
 * The class which performs the decoding process.
 * 
 * @author Amanda Askew & Dane Barney
 * @version 2003.09.11
 */
public class Decoder extends Endec {
    private BitReader in;

    /**
     * Opens the compressedFile for reading and interprets all information in the file
     * header (before the encoded data), creates a new arithmetic coder, and initializes
     * all the contexts.
     * 
     * @param compressedFile the file to decode
     */
    public Decoder(String compressedFile) throws InvalidFormatException, IOException {
        in = new BitReader(new DataInputStream(new FileInputStream(compressedFile)));
        ar = new ArithmeticCoder();

        // strip image header from file
        readImgHeader();

        // strip context coordinates from file
        readContextCoords();

        // create all the possible contexts and store them in an array
        createContextArray();
    }
    
    /**
     * Reads and interprets the image header data, which is stored in the compressed
     * file before the actual encoded pixel data.
     */
    public void readImgHeader() throws InvalidFormatException {
        try {
            int type = in.readByte(true);
            im = new Image(type, in, true);
            im.dominantColor = in.read4Bytes(true);
        } catch(IOException e) {
            // if an EOFException is thrown, the header is not properly formatted
            throw new InvalidFormatException("Invalid encoded file. Header is not recognized.");
        }
    }

    /**
     * Reads the positions of the pixels in the context that was used to encode this file.
     */
    public void readContextCoords() throws InvalidFormatException {
        try {
            contextCoords = new Coords[in.readByte(true)]; // first byte = total number of context coordinates
            for(int i=0; i < contextCoords.length; i++) {
                int x = in.readByte(true); // 1st byte = x coord
                int y = in.readByte(true); // 2nd byte = y coord
                contextCoords[i] = new Coords(x,y);
            }
        } catch(IOException e) {
            throw new InvalidFormatException("Invalid encoded file. Header is not recognized.");
        }
    }

    /**
     * Decodes pixels from the encoded data and returns the decoded Image object.
     */
    public Image decode() throws IOException {

        // prepare for decoding
        ar.initializeTag(in);

        // iterate through all the pixels in the pixel matrix. The decoder will know when to stop when the final
        // pixel is decoded
        for (int j=0; j < im.height; j++) {
            for (int i=0; i < im.width; i++) {
                // locate the correct context for this particular pixel
                Context context = findContext(i,j);

                // decode value and put in image
                im.pixels[i][j] = ar.decodeValue(context, in);
            }
        }
        
        in.close();
        
        return im;
    }
}